/*
    Theseus - maximum likelihood superpositioning of macromolecular structures

    Copyright (C) 2004-2015 Douglas L. Theobald

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the:

    Free Software Foundation, Inc.,
    59 Temple Place, Suite 330,
    Boston, MA  02111-1307  USA

    -/_|:|_|_\-
*/

#ifndef INTEGRATE_SEEN
#define INTEGRATE_SEEN

double
trapzd(double (*func)(double, double, double),
       double param1, double param2, double a, double b, int n);

double
integrate_qsimp(double (*func)(double, double, double),
                double param1, double param2, double a, double b);

double
integrate_romberg(double (*f)(double a, double p1, double p2),
                  double p1, double p2, double a, double b);

#endif
