/*
    Theseus - maximum likelihood superpositioning of macromolecular structures

    Copyright (C) 2004-2015 Douglas L. Theobald

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the:

    Free Software Foundation, Inc.,
    59 Temple Place, Suite 330,
    Boston, MA  02111-1307  USA

    -/_|:|_|_\-
*/

#ifndef MAT4UTILS_SEEN
#define MAT4UTILS_SEEN

void
Mat4Print(double **matrix);

void
Mat4Copy(double **matrix2, const double **matrix1);

void
Mat4TransposeIp(double **matrix);

void
Mat4TransposeOp(double **matrix2, const double **matrix1);

#endif /* !MATRIXUTILS_SEEN */ 
